#!/usr/bin/perl -w
#
# nntpswitch statistics generator
# $Id: nsstats.pl,v 1.26 2004/11/11 13:10:54 tommy Exp $
#
use strict;
use Time::ParseDate;
use POSIX qw(strftime mktime);
use Getopt::Std;

my $PrintUnknowns = 1;
my $TopStats	  = 30;
my $GroupTopStats = 50;
my $HTMLHeader    = "";
my $HTMLFooter    = "";
my $DIR;	  # set by option -d
my $FILE	  = undef;

my %opts;
my @unparsable  = ();
my @unknown     = ();
my %connects;	# connects by host
my $totconnects = 0;
my %commands;	# total commands by command
my $totcommands = 0;
my %articles;	# articles by host
my $totarticles = 0;
my %groups;	# groups by host
my $totgroups   = 0;
my %bytes;	# bytes by host
my $totbytes    = 0;
my %servers;	# server connects 
my %group;	# individual group requests 
my %timeouts;	# timeout strings
my %timeoutsl;  # timeout strings last time
my %sdowns;	# server downs
my %sdownsl;	# server downs last time
my %auths;	# authinfo results
my %grouparts;
my %groupbytes;
my %serverarts;
my %serverbytes;
my %hostuser;	# last authinfo user on host
my $totserverbytes = 0;
my $stime = 0;	# first log entry start time
my $etime = 0;  # and the end 
my %xtrace;	# shoud x-trace have a separate log?
my %postfilter;
my $totposts = 0;
my $totpostbytes = 0;
my %statindex;	# for index.html
my %posts;	# posts by host
my %postbytes;	# by host

# 
# fb() print number of bytes in human readable format
#
sub fb {
    my ($b) = @_;
    if ($b>=1000000000000) { return sprintf("%.2f Tb",$b/1000000000000); }
    if ($b>=1000000000) { return sprintf("%.2f Gb",$b/1000000000); }
    if ($b>=1000000) { return sprintf("%.2f Mb",$b/1000000); }
    if ($b>=1000) { return sprintf("%.2f Kb",$b/1000); }
    return $b;
}

sub hhead {
    if ( $HTMLHeader ne "" ) {
        open(IF,$HTMLHeader) || die $!;
	print $FILE <IF>;
	close(IF);
    }else{
        print $FILE <<STDHEADER;
	<HTML><HEAD><TITLE>NNTPSwitch Statistics</TITLE>
	<LINK REL="STYLESHEET" HREF="style.css">
	</HEAD>
	<BODY BGCOLOR=#FFFFFF>
STDHEADER
    }
}

sub hfoot {
    if ( $HTMLFooter ne "" ) {
        open(IF,$HTMLFooter) || die $!;
	print $FILE <IF>;
	close(IF);
    }else{
        print $FILE "</BODY></HTML>\n";
    }
}

############################################################
#
# pretty output routines, check for plain or html output
#
############################################################

sub htitle {
    my ($cs,$msg) = @_;
    print $FILE "<TABLE BORDER=1><TR><TH COLSPAN=$cs>$msg</TH></TR>\n";
}

sub hend {
    print $FILE "</TABLE><P>\n";
}

sub rowtab {
    print $FILE "<TR><TD>";
    print $FILE join("</TD><TD>",@_);
    print $FILE "</TD></TR>\n";
}

sub rowtab_right {
    print $FILE "<TR><TD ALIGN=RIGHT>";
    print $FILE join("</TD><TD ALIGN=RIGHT>",@_);
    print $FILE "</TD></TR>\n";
}

sub headtab {
    print $FILE "<TR><TH>";
    print $FILE join("</TH><TH>",@_);
    print $FILE "</TH></TR>\n";
}

sub ltgt {
    $_=~s/\</\&lt\;/g;
    $_=~s/\>/\&gt\;/g;
    $_;
}

############################################################
#
# read and parse the logfile from stdin
#
############################################################

getopts("d:", \%opts);
$DIR = $opts{d} ? $opts{d} : die("Usage: $0 -d /output/dir\n");

while(<STDIN>){
    chop;
    next if ! m/nntpswitch/;
    next if m/\: command /o;
    next if m/last message repeated/io;
    next if m/(read|write)(client|server)/io;
    next if m/too many concurrent connections/io;
    next if m/accounting_entry/io;
    next if m/connect to server/io;		# ??
    next if m/module.*loaded/io;
    next if m/active.*groups/io;
    next if m/active file freed/io;
    next if m/loaded.*acls/io;
    next if m/Freed.*(acls|Servers|Wildmats|Aliases|Authenticators)/io;
    next if m/loaded access.conf/io;
    next if m/Startup Test Successfull/io;
    next if m/disconnected from server (\S+)$/o;
    next if m/client error 500/io;
    next if m/post server responded (235|335)/io;
    next if m/received.*realmsgid/io;
    next if m/Session Time Limit Reached/o;
    next if m/Server running new pid/o;
    next if m/NNTP Server Starting/o;
    next if m/Listening on/o;
    next if m/updategroups\[/o;
    next if m/Loaded (\d+) (servers|newsgroups|overview.fmt|profiles|wildmats|aliases|authenticators)/i;
    next if m/Using configuration/o;
    next if m/Post received\,/;		# remove after migration
    next if m/DEBUG/;

    s/\^M//g;
    my ($month, $day, $time, $server, $daemon, $pid, undef) = 
      m/^(\S+)\s+(\d+)\s+([0-9\:]+)\s+(\S+)\s+(\S+)\[(\d+)\]\:(.*)/o;

    if ( ! $month || ! $day || ! $time || ! $server || ! $daemon || ! $pid ) {
        push @unparsable, $_;
	next;
    }

    $stime = parsedate("$month $day $time") if $stime == 0;
    $etime = parsedate("$month $day $time");

    if ( m/connect from (\S+)\:/o ) {
        next if ! $1;
	$connects{$1}++;
	$totconnects++;
	next;

    }elsif ( m/(\S+)\: user logged in (.*)/io ) {
        $hostuser{$1} = $2;
        next;

    }elsif ( m/(\S+)\: commands usage (.*)/o ) {
        my $cmds = $2;
    
        while ( $cmds =~ m/(\w+) (\d+) ?(.*)/ ) {
            $commands{uc($1)} += $2;
            $totcommands += $2;
            $cmds = $3;
        }
        next;

    }elsif ( m/\: group (\S+) closed articles (\d+) bytes (\d+)/io ) {
        next if ! $1;
	$grouparts{lc($1)} += $2;
	$groupbytes{lc($1)} += $3;
	next;

    }elsif ( m/\: group (\S+)/o ) {
        next if ! $1;
	$group{lc($1)}++;
	next;

    # remove 2nd m//, and defined $5 and $6 after migration
    }elsif ( m/(\S+)\: connection closed articles (\d+) bytes (\d+) groups (\d+) posts (\d+) postbytes (\d+)/o ||
    	     m/(\S+)\: connection closed articles (\d+) bytes (\d+) groups (\d+)/o ) {
	$articles{$1} += $2;
	$totarticles  += $2;
	$bytes{$1}    += $3;
	$totbytes     += $3;
	$groups{$1}   += $4;
	$totgroups    += $4;
	if ( defined $5 && defined $6 ) {
	    $posts{$1}    += $5;
	    $postbytes{$1}+= $6;
    	    $totposts     += $5;
	    $totpostbytes += $6;
        }
	next;

    }elsif ( m/disconnected from server (\S+) articles (\d+) bytes (\d+)/io ) {
        $serverarts{$1} += $2;
	$serverbytes{$1} += $3;
	$totserverbytes += $3;
	next;

    }elsif ( m/connecting to server (\S+)/o ) {
	$servers{$1}++;
	next;

    }elsif ( m/(backup server .*|remote server .*|cant s?write to server .*)/io ) {
    	$sdowns{$1}++;
	$sdownsl{$1} = $etime;
	next;

    }elsif ( m/(timeout after \d+ seconds|timeout or error .*|client read timeout|cant slwrite to client .*|cant write to server (.*)|eof from client|Can.t get default ACL|session (time|byte) limit reached|too many users|read_socket: select timeout|denied connection)/io ) {
	$timeouts{$1}++;
	$timeoutsl{$1} = $etime;
	next;

    }elsif ( m/Authentication Accepted/) {
	$auths{"Authentication Accepted"}++;
	next;

    }elsif ( m/authinfo result (.*)/o ) {
    	$auths{$1}++;
	next;

    }elsif ( m/X-Trace: (.*) (.*)/o ) {
        $xtrace{$1} = $_;
	next;

    }elsif ( m/post server responded (.*)/o ) {
    	$postfilter{$1}++;
	next;

    }elsif ( m/(Post Aborted or No body received|Post to nonexistent group)/o ) {
    	$postfilter{$1}++;
	next;

    }else{
	push @unknown, $_;
    }
}

############################################################
#
# print the report
#
############################################################

my $s = strftime("%Y%m%d_%H%M",localtime($stime));
my $e = strftime("%Y%m%d_%H%M",localtime($etime));
my $ss= strftime("%d-%m-%Y %H:%M:%S",localtime($stime));
my $ee= strftime("%d-%m-%Y %H:%M:%S",localtime($etime));

if ( ! open($FILE, ">$DIR/stats.$s-$e.html") ) {
    die("Can't open $DIR/stats.$s-$e.html for output, $!");
}

hhead();
print $FILE "<H1>NNTPSwitch Statistics $ss - $ee</H1>\n";

htitle(2,"Totals");
rowtab("Total Connects:",	    $totconnects);
rowtab("Total Commands:",	    $totcommands);
rowtab("Total Articles:",	    $totarticles);
rowtab("Total Accounted Bytes:",    fb($totbytes));
rowtab("Total Bytes from Server:",  fb($totserverbytes));
rowtab("Accounted Traffic:",	    sprintf("%.1f %%",((($totbytes+1)/($totserverbytes+1))*100)+.05));
rowtab("Total Group Changes:",	    $totgroups);
rowtab("Unique Groups:",	    scalar keys %group);
rowtab("Unique Hosts:",		    scalar keys %connects);
rowtab("Total Posts:",		    $totposts);
rowtab("Total Post Size:",	    fb($totpostbytes));
hend();

htitle(2,"Command usage");
foreach( sort { $commands{$b} <=> $commands{$a} } keys %commands ) {
    rowtab($_,$commands{$_});
}
hend();

htitle(4,"Server Connections");
headtab("Server", "Connections", "Articles", "Bytes");
foreach( sort keys %servers ) {
    $serverarts{$_} = 0 if ! $serverarts{$_};
    $serverbytes{$_} = 0 if ! $serverbytes{$_};
    rowtab($_, $servers{$_}, $serverarts{$_}, fb($serverbytes{$_}));
}
hend();

htitle(3,"Client Disconnect Reasons");
headtab("Last","Count","Message");
foreach( sort { $timeouts{$b} <=> $timeouts{$a} } keys %timeouts ) {
    rowtab(strftime("%H:%M:%S",localtime($timeoutsl{$_})), $timeouts{$_}, $_);
}
hend();

if ( keys %sdowns ) {
    htitle(3,"Server Downs");
    headtab("Last","Count","Message");
    foreach( sort { $sdowns{$b} <=> $sdowns{$a} } keys %sdowns ) {
        rowtab(strftime("%H:%M:%S",localtime($sdownsl{$_})), $sdowns{$_}, $_);
    }
    hend();
}

if ( keys %auths ) {
    htitle(2,"Authentication Results");
    foreach( sort { $auths{$b} <=> $auths{$a} } keys %auths ) {
        rowtab($_, $auths{$_});
    }
    hend();
}

if ( keys %postfilter ) {
    htitle(2,"Posting Filter Results");
    foreach( sort { $postfilter{$b} <=> $postfilter{$a} } keys %postfilter ) {
        rowtab($_, $postfilter{$_});
    }
    hend();
}

my $i = 0;
htitle(4,"Top $TopStats Posts by Host");
headtab("Host", "Posts", "Bytes", "User");
foreach( sort { $posts{$b} <=> $posts{$a} } keys %posts ) {
    last if $posts{$_} == 0;
    $hostuser{$_}  = "&nbsp;" if ! $hostuser{$_};
    $postbytes{$_} = 0   if ! $postbytes{$_};
    rowtab($_, $posts{$_}, fb($postbytes{$_}), $hostuser{$_} );
    $i++;
    last if $i >= $TopStats;
}
hend();

$i=0;
htitle(4,"Top $TopStats Post Bytes by Host");
headtab("Host", "Posts", "Bytes", "User");
foreach( sort { $postbytes{$b} <=> $postbytes{$a} } keys %postbytes ) {
    last if $postbytes{$_} == 0;
    $hostuser{$_}  = "&nbsp;" if ! $hostuser{$_};
    $postbytes{$_} = 0   if ! $postbytes{$_};
    rowtab($_, $posts{$_}, fb($postbytes{$_}), $hostuser{$_} );
    $i++;
    last if $i >= $TopStats;
}
hend();

$i = 0;
htitle(6,"Top $TopStats Connects by Host");
headtab("Host", "Connects", "Articles", "Groups", "Bytes", "User");
foreach( sort { $connects{$b} <=> $connects{$a} } keys %connects ) {
    last if $connects{$_} == 0;
    $hostuser{$_}="&nbsp;" if ! $hostuser{$_};
    $connects{$_}=0 if ! $connects{$_};
    $articles{$_}=0 if ! $articles{$_};
    $groups{$_}=0   if ! $groups{$_};
    $bytes{$_}=0    if ! $bytes{$_};
    rowtab($_, $connects{$_}, $articles{$_}, $groups{$_}, fb($bytes{$_}), $hostuser{$_} );
    $i++;
    last if $i >= $TopStats;
}
hend();

$i = 0;
htitle(6,"Top $TopStats Bytes by Host");
headtab("Host", "Connects", "Articles", "Groups", "Bytes", "User");
foreach( sort { $bytes{$b} <=> $bytes{$a} } keys %bytes ) {
    last if $bytes{$_} == 0;
    $hostuser{$_}="&nbsp;" if ! $hostuser{$_};
    $connects{$_}=0 if ! $connects{$_};
    $articles{$_}=0 if ! $articles{$_};
    $groups{$_}=0   if ! $groups{$_};
    rowtab($_, $connects{$_}, $articles{$_}, $groups{$_}, fb($bytes{$_}), $hostuser{$_} );
    $i++;
    last if $i >= $TopStats;
}
hend();


$i = 0;
htitle(6,"Top $TopStats Groups by Host");
headtab("Host", "Connects", "Articles", "Groups", "Bytes", "User");
foreach( sort { $groups{$b} <=> $groups{$a} } keys %groups ) {
    last if $groups{$_} == 0;
    $hostuser{$_}="&nbsp;" if ! $hostuser{$_};
    $connects{$_}=0 if ! $connects{$_};
    rowtab($_, $connects{$_}, $articles{$_}, $groups{$_}, fb($bytes{$_}), $hostuser{$_} );
    $i++;
    last if $i >= $TopStats;
}
hend();


$i = 0;
htitle(6,"Top $TopStats Articles by Host");
headtab("Host", "Connects", "Articles", "Groups", "Bytes", "User");
foreach( sort { $articles{$b} <=> $articles{$a} } keys %articles ) {
    last if $articles{$_} == 0;
    $hostuser{$_}="unknown" if ! $hostuser{$_};
    $connects{$_}=0 if ! $connects{$_};
    rowtab($_, $connects{$_}, $articles{$_}, $groups{$_}, fb($bytes{$_}), $hostuser{$_} );
    $i++;
    last if $i >= $TopStats;
}
hend();


$i = 0;
htitle(4,"Top $GroupTopStats Group Statistics by Article");
headtab("Group", "Connects", "Articles", "Bytes");
foreach( sort { $grouparts{$b} <=> $grouparts{$a} } keys %grouparts ) {
    rowtab($_, $group{$_}, $grouparts{$_}, fb($groupbytes{$_}));
    $i++;
    last if $i >= $GroupTopStats;
}
hend();

$i = 0;
htitle(4,"Top $GroupTopStats Group Statistics by Bytes");
headtab("Group", "Connects", "Articles", "Bytes");
foreach( sort { $groupbytes{$b} <=> $groupbytes{$a} } keys %groupbytes ) {
    rowtab($_, $group{$_}, $grouparts{$_}, fb($groupbytes{$_}));
    $i++;
    last if $i >= $GroupTopStats;
}
hend();

if ( $PrintUnknowns ) {
    if ( @unparsable ) {
        htitle(1,"Unparsable Log Entries");
        for(@unparsable){
            rowtab(ltgt($_));
        }
	hend();
    }

    if ( @unknown ) {
	htitle(1,"Unknown Log Entries");
	for(@unknown){
            rowtab(ltgt($_));
	}
	hend();
    }
}
hfoot();
close($FILE);

############################################################
#
# create the trace log
#
############################################################

my ($trace, $uhost, $msgid);

if ( $totposts > 0 ) {

if ( ! open($FILE, ">$DIR/trace.$s-$e.txt") ) {
    die("Can't open $DIR/trace.$s-$e.txt for output, $!");
}

print $FILE "NNTPSwitch X-Trace Log $ss - $ee\n\n";
print $FILE "X-Trace                    Username           Hostname                                 Message-ID\n";

foreach( keys %xtrace ) {
    ($trace,$uhost,$msgid) = split(/ /,$_);
    if ( $uhost =~ m/\@/ ) {
        $uhost =~ m/(.*)\@(.*)$/;
        print $FILE sprintf("%-26s %-18s %-40.40s %s\n", $trace, $1, $2, $msgid);
    }else{
        print $FILE sprintf("%-26s %-18s %-40.40s %s\n", $trace, " ", $uhost, $msgid);
    }
}

close $FILE;
} #totposts>0

############################################################
#
# create the index.html 
#
############################################################

my $d;
if ( ! opendir($d,$DIR) ) { die "Cant open $DIR $!\n"; }
while($_=readdir($d)) {
   next if ! m/^stats\.([0-9_]+)\-([0-9_]+)\.html$/o;
															       my ($y,$m,$d,$x,$h,$min) = unpack("A4A2A2AA2A2",$1);
   my $tm =  mktime(0, $min, $h, $d, $m-1, $y-1900, 0, 0, 0);
   $statindex{$tm} = "$1-$2";
}
closedir($d);

if ( ! open($FILE, ">$DIR/index.html") ) {
    die("Can't open $DIR/index.html for output, $!");
}

hhead();
print $FILE "<H1>NNTPSwitch Statistics</H1>\n";
htitle(10,"Daily Statistics");
headtab("Date", "Connections", "Commands", "Articles", "Bytes", "Groups", "Hosts", "Posts", "Post Size", "Log");

foreach( sort { $b cmp $a } keys %statindex ) {
    my $dat = $_;
    my $f   = sprintf("stats.%s.html",$statindex{$dat});
    my $tf  = sprintf("trace.%s.txt", $statindex{$dat});
    if ( -f "$DIR/$tf" ) {
        $tf = "<a href=$tf>trace</a>";
    }else{
        $tf = "n/a";
    }

    my ($tc, $tcm, $ta, $tb, $tug, $tuh, $ps, $pb) = (0,0,0,0,0,0,0,0);

    if ( open(F, "$DIR/$f") ) {
	while(<F>) {
	    if ( m/Total Connects\:\<\/TD\>\<TD\>(\d+)/ ) {
		$tc = $1;
	    }elsif ( m/Total Commands\:\<\/TD\>\<TD\>(\d+)/ ) {
		$tcm = $1;
	    }elsif ( m/Total Articles\:\<\/TD\>\<TD\>(\d+)/ ) {
		$ta = $1;
	    }elsif ( m/Total Accounted Bytes\:\<\/TD\>\<TD\>(.*)\<\/TD>/ ) {
		$tb = $1;
	    }elsif ( m/Unique Groups\:\<\/TD\>\<TD\>(\d+)/ ) {
		$tug = $1;
	    }elsif ( m/Unique Hosts\:\<\/TD\>\<TD\>(\d+)/ ) {
		$tuh = $1;
	    }elsif ( m/Total Posts\:\<\/TD\>\<TD\>(\d+)/ ) {
		$ps = $1;
	    }elsif ( m/Total Post Size\:\<\/TD\>\<TD\>(.*)\<\/TD>/ ) {
		$pb = $1;
	    }
	    last if m/Command Usage/;
	}
	close(F);

	my $date = strftime("<a href=$f>%d-%b-%Y %H:%M</a>",localtime($dat));
	rowtab_right($date, $tc, $tcm, $ta, $tb, $tug, $tuh, $ps, $pb, $tf);
    }
}
hend();
hfoot();
close $FILE;

